/*
 * \file polimorf.cpp
 *
 * Többszörös öröklés és a polimorf viselkedés,
 * valamint a dynamic_cast memutatása.
 *
 * A dynamic_cast-hoz fontos, hogy a fordításkor legyen
 * engedélyezve a Run-Time Type Info (RTTI) generálása
 * (Vis.stud: /GR opció)
 * GCC-nél automatikus
 *
 */

#include <iostream>
#include <stdexcept>

using namespace std;

/// A osztáy.
/// (struct: mindene publikus)
struct A {
    void f()        { cout << " A::f()\n"; }
    void f(int)     { cout << " A::f(int)\n";  }
    virtual ~A() {} // csak ezért virt., hogy legyen RTTI infó
};

/// B osztáy.
/// (struct: mindene publikus)
struct B {
    void f(double)  { cout << " B::f(double)\n"; };
};


/// C osztáy.
/// (struct: mindene publikus)
struct C {
    void f(int*)  { cout << " C::f(int*)\n"; };
};

/// AB osztáy.
class AB :public A, public B {
public:
    // Többszörös öröklésnél az azonos nevű tagfüggvényekről a fordító feltételezi,
    // hogy teljesen eltérő viselkedést valósítanak meg, ezért egyik elérését sem
    // engedi automatikusan, hiába publikus az öröklés.
    using A::f;     // Így elérhetõk az A-beli f()-ek
    using B::f;     // Így elérhető a B-beli f()is
    void f() {
        cout << " AB::f()\n";
        A::f();
        f(1);       // A::f(int)
        f(1.2);     // B::f(double)
    }
};

/// Kiprobáljuk...
int main() {
    AB ab;

    cout << "Fuggvenyek [ab.A::f(), ab.f(int) ab.f(doube) ab.f()]" << endl;
    ab.A::f();
    ab.f(5);        // A::f(int)
    ab.f(6.3);      // B::f(double)
    ab.f();         // AB::f()

    A *Ap = &ab;    // az upcast automaiikus
    B *Bp = &ab;    // az upcast automaiikus

    cout << endl << "Pointerek konvertalasa:" << endl;
    AB *ABp = dynamic_cast<AB*>(Ap);    // biztonságos down cast
    if (ABp != 0) {
        cout << "Sikeres volt a down cast" << endl;
        ABp->f();
    } else {
        cout << "A down cast nem sikerult" << endl;
    }

    C *Cp = dynamic_cast<C*>(Ap);        // ez nem fog menni
    if (Cp == 0) cout << "AB* nem konvertalhato C*-ra" << endl;

    cout << endl << "Referencia konvertalasa:" << endl;
    // Referencia konvertálásánál a hibajelzés kivétellel jön
    try {
        AB& ab2 = dynamic_cast<AB&>(*Ap);
        cout << "Sikeres volt a down cast referenciara is:" << endl;
        ab2.f();
        cout << "Kiserlet C referenciara valo konvertziora:" << endl;
        C& c2 = dynamic_cast<C&>(*Ap);
    } catch (exception& e) {
        cout << e.what() << endl;
    }
    return 0;
}
